// (C) 2001-2015 Altera Corporation. All rights reserved.
// Your use of Altera Corporation's design tools, logic functions and other 
// software and tools, and its AMPP partner logic functions, and any output 
// files any of the foregoing (including device programming or simulation 
// files), and any associated documentation or information are expressly subject 
// to the terms and conditions of the Altera Program License Subscription 
// Agreement, Altera MegaCore Function License Agreement, or other applicable 
// license agreement, including, without limitation, that your use is for the 
// sole purpose of programming logic devices manufactured by Altera and sold by 
// Altera or its authorized distributors.  Please refer to the applicable 
// agreement for further details.


//////////////////////////////////////////////////////////////////////////////
// When enabled, the read compare module buffers the write data and compares
// it with the returned read data.  If the write and read data do not match,
// the corresponding bits of pnf_per_bit is deasserted.
//////////////////////////////////////////////////////////////////////////////
module altera_emif_avl_tg_read_compare # (
   parameter DEVICE_FAMILY            = "",
   parameter DATA_WIDTH               = "",
   parameter BE_WIDTH                 = "",
   parameter WRITTEN_DATA_FIFO_SIZE   = "",
   parameter TG_LFSR_SEED             = 36'b000000111110000011110000111000110010,
   parameter TG_RANDOM_BYTE_ENABLE    = 1
) (
   // Clock and reset
   input                          clk,
   input                          reset_n,

   // Control signals
   input                          enable,
   input                          queue_read_compare_req,

   // Avalon read data
   input                          rdata_valid,
   input logic [DATA_WIDTH-1:0]   rdata,

   // Read compare status
   output logic                   read_compare_fifo_full,
   output logic                   read_compare_fifo_empty,
   output logic [DATA_WIDTH-1:0]  pnf_per_bit
);
   timeunit 1ns;
   timeprecision 1ps;
   
   import avl_tg_defs::*;

   // Byte size derived from dividing data width by byte enable width
   // Round up so that compile fails if DATA_WIDTH is not a multiple of BE_WIDTH
   localparam BYTE_SIZE = (DATA_WIDTH + BE_WIDTH - 1) / BE_WIDTH;

   // the width of the local data counter
   localparam DATACOUNTER_WIDTH = 8;
   
   // width of counter to emulate a fifo of read comparison requests
   localparam WRITTEN_DATA_COUNTER_WIDTH = ceil_log2(WRITTEN_DATA_FIFO_SIZE) + 2;

   // Write data
   logic [DATA_WIDTH-1:0]        written_data;
   logic [DATA_WIDTH-1:0]        written_data_lfsr_out;
   logic [BE_WIDTH-1:0]          written_be_lfsr_out;

   // Read/write data registers
   logic                         rdata_valid_r;
   logic                         rdata_valid_rr;
   logic [DATA_WIDTH-1:0]        rdata_r;
   logic [DATA_WIDTH-1:0]        rdata_rr;
   logic                         queue_read_compare_req_r;
   logic                         queue_read_compare_req_rr;

   // Data Counter
   logic [DATACOUNTER_WIDTH-1:0] data_counter;
   
   // Counter to emulate a fifo of read comparison requests
   logic [WRITTEN_DATA_COUNTER_WIDTH-1:0] written_data_counter;

   // Should errors be forced?
   logic                         force_error;
   
   // Enable LFSR to re-generate next expected write data/data mask
   logic                         enable_lfsr;

   // Per bit comparison
   always_ff @(posedge clk or negedge reset_n)
   begin
      if (!reset_n) begin
         pnf_per_bit <= {DATA_WIDTH{1'b1}};
      end else begin
         for (int byte_num = 0; byte_num < BE_WIDTH; byte_num++) begin
            for (int bit_num = byte_num * BYTE_SIZE; bit_num < (byte_num + 1) * BYTE_SIZE; bit_num++) begin
               if (enable && rdata_valid_rr && written_be_lfsr_out[byte_num])
                  pnf_per_bit[bit_num] <= (rdata_rr[bit_num] === written_data[bit_num]);
               else
                  pnf_per_bit[bit_num] <= 1'b1;
            end
         end
      end
   end

   always_ff @(posedge clk or negedge reset_n)
   begin
      if (!reset_n) begin
         rdata_valid_r             <= 1'b0;
         rdata_valid_rr            <= 1'b0;
         queue_read_compare_req_r  <= 1'b0;
         queue_read_compare_req_rr <= 1'b0;
      end
      else
      begin
         rdata_valid_r             <= rdata_valid;
         rdata_valid_rr            <= rdata_valid_r;
         queue_read_compare_req_r  <= queue_read_compare_req;
         queue_read_compare_req_rr <= queue_read_compare_req_r;
      end
   end
   
   always_ff @(posedge clk)
   begin
      rdata_r  <= rdata;
      rdata_rr <= rdata_r;
   end   

   // The data is used as a small counter to count data coming back. It is
   // used by the force_error mode to introduce errors.
   always_ff @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
         data_counter <= '0;
      else
         if (rdata_valid_rr)
            data_counter <= data_counter + 1'b1;
   end

   // synthesis translate_off   
   // Display a message to the user if there is an error
   logic [DATA_WIDTH-1:0]   rdata_rrr;
   logic [DATA_WIDTH-1:0]   written_data_rr;
   logic [DATA_WIDTH-1:0]   written_be_full_rr;
   logic [DATA_WIDTH-1:0]   written_be_full;

   // Generate bit-wise byte-enable signal which is easier to read
   generate
   genvar byte_num;
      for (byte_num = 0; byte_num < BE_WIDTH; ++byte_num) 
      begin : gen_written_be_full
         assign written_be_full [byte_num * BYTE_SIZE +: BYTE_SIZE] = {BYTE_SIZE{written_be_lfsr_out[byte_num]}};
      end
   endgenerate

   always_ff @(posedge clk)
   begin
      rdata_rrr <= rdata_rr;
      written_data_rr <= written_data;
      written_be_full_rr <= written_be_full;
      
      if (~(&pnf_per_bit))
      begin
         $display("[%0t] ERROR: Expected %h/%h but read %h", $time, written_data_rr, written_be_full_rr, rdata_rrr);
         $display("            wrote bits: %h", written_data_rr & written_be_full_rr);
         $display("             read bits: %h", rdata_rrr & written_be_full_rr);
      end
   end
   // synthesis translate_on

   assign force_error = 1'b0;

   assign written_data = (force_error) ?
      ((data_counter > 10) ? {written_data_lfsr_out[DATA_WIDTH-1:1],~written_data_lfsr_out[0]} : written_data_lfsr_out) :
      written_data_lfsr_out;
   
   // We use LFSRs to re-generate the random write data/mask for 
   // read comparison. By using the same seed we're guaranteed
   // to regenerate the same sequence. This saves us from the 
   // need of instantiating a FIFO to record the write data (and
   // the timing closure challenge associated with it), at
   // the expense of flexibility (that we currently don't need).
   // For example if one of the stages need to use special data
   // instead of LFSR, we can't handle it.
   
   // Enable LFSR to generate the next item
   assign enable_lfsr = enable & rdata_valid_rr;
   
   // Write data generator
   altera_emif_avl_tg_lfsr_wrapper # (
      .DATA_WIDTH   (DATA_WIDTH),
      .SEED         (TG_LFSR_SEED)
   ) data_gen_inst (
      .clk          (clk),
      .reset_n      (reset_n),
      .enable       (enable_lfsr),
      .data         (written_data_lfsr_out)
   );
   
   // Byte enable generator
   generate
      if (TG_RANDOM_BYTE_ENABLE == 1)
      begin : be_gen
         altera_emif_avl_tg_lfsr_wrapper # (
            .DATA_WIDTH   (BE_WIDTH)
         ) be_gen_inst (
            .clk          (clk),
            .reset_n      (reset_n),
            .enable       (enable_lfsr),
            .data         (written_be_lfsr_out)
         );
      end 
      else 
      begin : be_const
         assign written_be_lfsr_out = '1;
      end
   endgenerate      

   always_ff @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
         written_data_counter <= '1;
      else if (enable & queue_read_compare_req_rr & rdata_valid_r)
         written_data_counter <= written_data_counter;
      else if (enable & queue_read_compare_req_rr)
         written_data_counter <= written_data_counter + 1'b1;
      else if (enable & rdata_valid_r)
         written_data_counter <= written_data_counter - 1'b1;
      else 
         written_data_counter <= written_data_counter;
   end
   
   assign read_compare_fifo_full  = (written_data_counter[WRITTEN_DATA_COUNTER_WIDTH-1:WRITTEN_DATA_COUNTER_WIDTH-2] == 2'b01);
   assign read_compare_fifo_empty = written_data_counter[WRITTEN_DATA_COUNTER_WIDTH-1];
   
endmodule

